@@ -21,11 +21,15 @@ import (
 	"encoding/json"
 	"fmt"
 	"net"
+	"os"
 	"reflect"
+	"strconv"
 	"time"
 
 	"k8s.io/apimachinery/pkg/api/equality"
+	"k8s.io/client-go/tools/record"
 
+	clusterv1 "github.com/openshift/cluster-api/pkg/apis/cluster/v1alpha1"
 	machinev1 "github.com/openshift/cluster-api/pkg/apis/machine/v1beta1"
 	apierrors "github.com/openshift/cluster-api/pkg/errors"
 	"github.com/openshift/cluster-api/pkg/util"
@@ -39,6 +43,7 @@ import (
 	"sigs.k8s.io/cluster-api-provider-openstack/pkg/bootstrap"
 	"sigs.k8s.io/cluster-api-provider-openstack/pkg/cloud/openstack"
 	"sigs.k8s.io/cluster-api-provider-openstack/pkg/cloud/openstack/clients"
+	"sigs.k8s.io/cluster-api-provider-openstack/pkg/cloud/openstack/options"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 
 	clconfig "github.com/coreos/container-linux-config-transpiler/config"
@@ -51,11 +56,23 @@ const (
 	DisableTemplatingKey = "disableTemplating"
 	PostprocessorKey     = "postprocessor"
 
-	TimeoutInstanceCreate       = 5 * time.Minute
-	TimeoutInstanceDelete       = 5 * time.Minute
+	TimeoutInstanceCreate       = 5
+	TimeoutInstanceDelete       = 5
 	RetryIntervalInstanceStatus = 10 * time.Second
 
-	TokenTTL = 60 * time.Minute
+	// MachineInstanceStateAnnotationName as annotation name for a machine instance state
+	MachineInstanceStateAnnotationName = "machine.openshift.io/instance-state"
+
+	// ErrorState is assigned to the machine if its instance has been destroyed
+	ErrorState = "ERROR"
+)
+
+// Event Action Constants
+const (
+	createEventAction = "Create"
+	updateEventAction = "Update"
+	deleteEventAction = "Delete"
+	noEventAction     = ""
 )
 
 type OpenstackClient struct {
@@ -63,6 +80,7 @@ type OpenstackClient struct {
 	scheme *runtime.Scheme
 	client client.Client
 	*openstack.DeploymentClient
+	eventRecorder record.EventRecorder
 }
 
 func NewActuator(params openstack.ActuatorParams) (*OpenstackClient, error) {
@@ -71,10 +89,38 @@ func NewActuator(params openstack.ActuatorParams) (*OpenstackClient, error) {
 		client:           params.Client,
 		scheme:           params.Scheme,
 		DeploymentClient: openstack.NewDeploymentClient(),
+		eventRecorder:    params.EventRecorder,
 	}, nil
 }
 
-func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluster, machine *machinev1.Machine) error {
+func getTimeout(name string, timeout int) time.Duration {
+	if v := os.Getenv(name); v != "" {
+		timeout, err := strconv.Atoi(v)
+		if err == nil {
+			return time.Duration(timeout)
+		}
+	}
+	return time.Duration(timeout)
+}
+
+func (oc *OpenstackClient) Create(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
+	// First check that provided labels are correct
+	// TODO(mfedosin): stop sending the infrastructure request when we start to receive the cluster value
+	clusterInfra, err := oc.params.ConfigClient.Infrastructures().Get("cluster", metav1.GetOptions{})
+	if err != nil {
+		return fmt.Errorf("Failed to retrieve cluster Infrastructure object: %v", err)
+	}
+
+	clusterInfraName := clusterInfra.Status.InfrastructureName
+	clusterNameLabel := machine.Labels["machine.openshift.io/cluster-api-cluster"]
+
+	if clusterNameLabel != clusterInfraName {
+		klog.Errorf("machine.openshift.io/cluster-api-cluster label value is incorrect: %v, machine %v cannot join cluster %v", clusterNameLabel, machine.ObjectMeta.Name, clusterInfraName)
+		verr := apierrors.InvalidMachineConfiguration("machine.openshift.io/cluster-api-cluster label value is incorrect: %v, machine %v cannot join cluster %v", clusterNameLabel, machine.ObjectMeta.Name, clusterInfraName)
+
+		return oc.handleMachineError(machine, verr, createEventAction)
+	}
+
 	kubeClient := oc.params.KubeClient
 
 	machineService, err := clients.NewInstanceServiceFromMachine(kubeClient, machine)
@@ -85,11 +131,11 @@ func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluste
 	providerSpec, err := openstackconfigv1.MachineSpecFromProviderSpec(machine.Spec.ProviderSpec)
 	if err != nil {
 		return oc.handleMachineError(machine, apierrors.InvalidMachineConfiguration(
-			"Cannot unmarshal providerSpec field: %v", err))
+			"Cannot unmarshal providerSpec field: %v", err), createEventAction)
 	}
 
 	if verr := oc.validateMachine(machine, providerSpec); verr != nil {
-		return oc.handleMachineError(machine, verr)
+		return oc.handleMachineError(machine, verr, createEventAction)
 	}
 
 	instance, err := oc.instanceExists(machine)
@@ -101,6 +147,18 @@ func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluste
 		return nil
 	}
 
+	// Here we check whether we want to create a new instance or recreate the destroyed
+	// one. If this is the second case, we have to return an error, because if we just
+	// create an instance with the old name, because the CSR for it will not be approved
+	// automatically.
+	// See https://bugzilla.redhat.com/show_bug.cgi?id=1746369
+	if machine.ObjectMeta.Annotations[InstanceStatusAnnotationKey] != "" {
+		klog.Errorf("The instance has been destroyed for the machine %v, cannot recreate it.\n", machine.ObjectMeta.Name)
+		verr := apierrors.InvalidMachineConfiguration("the instance has been destroyed for the machine %v, cannot recreate it.\n", machine.ObjectMeta.Name)
+
+		return oc.handleMachineError(machine, verr, createEventAction)
+	}
+
 	// get machine startup script
 	var ok bool
 	var disableTemplating bool
@@ -138,23 +196,24 @@ func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluste
 
 	var userDataRendered string
 	if len(userData) > 0 && !disableTemplating {
-		if machine.Spec.Versions.ControlPlane != "" {
+		// FIXME(mandre) Find the right way to check if machine is part of the control plane
+		if machine.ObjectMeta.Name != "" {
 			userDataRendered, err = masterStartupScript(cluster, machine, string(userData))
 			if err != nil {
 				return oc.handleMachineError(machine, apierrors.CreateMachine(
-					"error creating Openstack instance: %v", err))
+					"error creating Openstack instance: %v", err), createEventAction)
 			}
 		} else {
 			klog.Info("Creating bootstrap token")
 			token, err := oc.createBootstrapToken()
 			if err != nil {
 				return oc.handleMachineError(machine, apierrors.CreateMachine(
-					"error creating Openstack instance: %v", err))
+					"error creating Openstack instance: %v", err), createEventAction)
 			}
 			userDataRendered, err = nodeStartupScript(cluster, machine, token, string(userData))
 			if err != nil {
 				return oc.handleMachineError(machine, apierrors.CreateMachine(
-					"error creating Openstack instance: %v", err))
+					"error creating Openstack instance: %v", err), createEventAction)
 			}
 		}
 	} else {
@@ -202,14 +261,15 @@ func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluste
 		}
 	}
 
-	instance, err = machineService.InstanceCreate(clusterName, machine.Name, &clusterSpec, providerSpec, userDataRendered, providerSpec.KeyName)
+	instance, err = machineService.InstanceCreate(clusterName, machine.Name, &clusterSpec, providerSpec, userDataRendered, providerSpec.KeyName, oc.params.ConfigClient)
 
 	if err != nil {
 		return oc.handleMachineError(machine, apierrors.CreateMachine(
-			"error creating Openstack instance: %v", err))
+			"error creating Openstack instance: %v", err), createEventAction)
 	}
-	// TODO: wait instance ready
-	err = util.PollImmediate(RetryIntervalInstanceStatus, TimeoutInstanceCreate, func() (bool, error) {
+	instanceCreateTimeout := getTimeout("CLUSTER_API_OPENSTACK_INSTANCE_CREATE_TIMEOUT", TimeoutInstanceCreate)
+	instanceCreateTimeout = instanceCreateTimeout * time.Minute
+	err = util.PollImmediate(RetryIntervalInstanceStatus, instanceCreateTimeout, func() (bool, error) {
 		instance, err := machineService.GetInstance(instance.ID)
 		if err != nil {
 			return false, nil
@@ -218,22 +278,28 @@ func (oc *OpenstackClient) Create(ctx context.Context, cluster *machinev1.Cluste
 	})
 	if err != nil {
 		return oc.handleMachineError(machine, apierrors.CreateMachine(
-			"error creating Openstack instance: %v", err))
+			"error creating Openstack instance: %v", err), createEventAction)
 	}
 
 	if providerSpec.FloatingIP != "" {
 		err := machineService.AssociateFloatingIP(instance.ID, providerSpec.FloatingIP)
 		if err != nil {
 			return oc.handleMachineError(machine, apierrors.CreateMachine(
-				"Associate floatingIP err: %v", err))
+				"Associate floatingIP err: %v", err), createEventAction)
 		}
 
 	}
 
+	err = machineService.SetMachineLabels(machine, instance.ID)
+	if err != nil {
+		return nil
+	}
+
+	oc.eventRecorder.Eventf(machine, corev1.EventTypeNormal, "Created", "Created Machine %v", machine.Name)
 	return oc.updateAnnotation(machine, instance.ID)
 }
 
-func (oc *OpenstackClient) Delete(ctx context.Context, cluster *machinev1.Cluster, machine *machinev1.Machine) error {
+func (oc *OpenstackClient) Delete(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
 	machineService, err := clients.NewInstanceServiceFromMachine(oc.params.KubeClient, machine)
 	if err != nil {
 		return err
@@ -253,13 +319,14 @@ func (oc *OpenstackClient) Delete(ctx context.Context, cluster *machinev1.Cluste
 	err = machineService.InstanceDelete(id)
 	if err != nil {
 		return oc.handleMachineError(machine, apierrors.DeleteMachine(
-			"error deleting Openstack instance: %v", err))
+			"error deleting Openstack instance: %v", err), deleteEventAction)
 	}
 
+	oc.eventRecorder.Eventf(machine, corev1.EventTypeNormal, "Deleted", "Deleted machine %v", machine.Name)
 	return nil
 }
 
-func (oc *OpenstackClient) Update(ctx context.Context, cluster *machinev1.Cluster, machine *machinev1.Machine) error {
+func (oc *OpenstackClient) Update(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
 	status, err := oc.instanceStatus(machine)
 	if err != nil {
 		return err
@@ -273,6 +340,18 @@ func (oc *OpenstackClient) Update(ctx context.Context, cluster *machinev1.Cluste
 		}
 		if instance != nil && instance.Status == "ACTIVE" {
 			klog.Infof("Populating current state for boostrap machine %v", machine.ObjectMeta.Name)
+
+			kubeClient := oc.params.KubeClient
+			machineService, err := clients.NewInstanceServiceFromMachine(kubeClient, machine)
+			if err != nil {
+				return err
+			}
+
+			err = machineService.SetMachineLabels(machine, instance.ID)
+			if err != nil {
+				return nil
+			}
+
 			return oc.updateAnnotation(machine, instance.ID)
 		} else {
 			return fmt.Errorf("Cannot retrieve current state to update machine %v", machine.ObjectMeta.Name)
@@ -283,7 +362,8 @@ func (oc *OpenstackClient) Update(ctx context.Context, cluster *machinev1.Cluste
 		return nil
 	}
 
-	if currentMachine.Spec.Versions.ControlPlane != "" {
+	// FIXME(mandre) Find the right way to check if machine is part of the control plane
+	if currentMachine.ObjectMeta.Name != "" {
 		// TODO: add master inplace
 		klog.Errorf("master inplace update failed: not support master in place update now")
 	} else {
@@ -291,34 +371,37 @@ func (oc *OpenstackClient) Update(ctx context.Context, cluster *machinev1.Cluste
 		err = oc.Delete(ctx, cluster, currentMachine)
 		if err != nil {
 			klog.Errorf("delete machine %s for update failed: %v", currentMachine.ObjectMeta.Name, err)
-		} else {
-			err = util.PollImmediate(RetryIntervalInstanceStatus, TimeoutInstanceDelete, func() (bool, error) {
-				instance, err := oc.instanceExists(machine)
-				if err != nil {
-					return false, nil
-				}
-				return instance == nil, nil
-			})
-			if err != nil {
-				return oc.handleMachineError(machine, apierrors.DeleteMachine(
-					"error deleting Openstack instance: %v", err))
-			}
-
-			err = oc.Create(ctx, cluster, machine)
-			if err != nil {
-				klog.Errorf("create machine %s for update failed: %v", machine.ObjectMeta.Name, err)
-			}
-			klog.Infof("Successfully updated machine %s", currentMachine.ObjectMeta.Name)
+			return fmt.Errorf("Cannot delete machine %s: %v", currentMachine.ObjectMeta.Name, err)
 		}
+		instanceDeleteTimeout := getTimeout("CLUSTER_API_OPENSTACK_INSTANCE_DELETE_TIMEOUT", TimeoutInstanceDelete)
+		instanceDeleteTimeout = instanceDeleteTimeout * time.Minute
+		err = util.PollImmediate(RetryIntervalInstanceStatus, instanceDeleteTimeout, func() (bool, error) {
+			instance, err := oc.instanceExists(machine)
+			if err != nil {
+				return false, nil
+			}
+			return instance == nil, nil
+		})
+		if err != nil {
+			return oc.handleMachineError(machine, apierrors.DeleteMachine(
+				"error deleting Openstack instance: %v", err), deleteEventAction)
+		}
+		err = oc.Create(ctx, cluster, machine)
+		if err != nil {
+			klog.Errorf("create machine %s for update failed: %v", machine.ObjectMeta.Name, err)
+			return fmt.Errorf("Cannot create machine %s: %v", machine.ObjectMeta.Name, err)
+		}
+		klog.Infof("Successfully updated machine %s", currentMachine.ObjectMeta.Name)
 	}
 
+	oc.eventRecorder.Eventf(currentMachine, corev1.EventTypeNormal, "Updated", "Updated machine %v", currentMachine.ObjectMeta.Name, updateEventAction)
 	return nil
 }
 
-func (oc *OpenstackClient) Exists(ctx context.Context, cluster *machinev1.Cluster, machine *machinev1.Machine) (bool, error) {
+func (oc *OpenstackClient) Exists(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) (bool, error) {
 	instance, err := oc.instanceExists(machine)
 	if err != nil {
-		return false, err
+		return false, fmt.Errorf("Error checking if instance exists (machine/actuator.go 346): %v", err)
 	}
 	return instance != nil, err
 }
@@ -363,12 +446,22 @@ func getIPFromInstance(instance *clients.Instance) (string, error) {
 // the appropriate reason/message on the Machine.Status. If not, such as during
 // cluster installation, it will operate as a no-op. It also returns the
 // original error for convenience, so callers can do "return handleMachineError(...)".
-func (oc *OpenstackClient) handleMachineError(machine *machinev1.Machine, err *apierrors.MachineError) error {
+func (oc *OpenstackClient) handleMachineError(machine *machinev1.Machine, err *apierrors.MachineError, eventAction string) error {
+	if eventAction != noEventAction {
+		oc.eventRecorder.Eventf(machine, corev1.EventTypeWarning, "Failed"+eventAction, "%v", err.Reason)
+	}
 	if oc.client != nil {
 		reason := err.Reason
 		message := err.Message
 		machine.Status.ErrorReason = &reason
 		machine.Status.ErrorMessage = &message
+
+		// Set state label to indicate that this machine is broken
+		if machine.ObjectMeta.Annotations == nil {
+			machine.ObjectMeta.Annotations = make(map[string]string)
+		}
+		machine.ObjectMeta.Annotations[MachineInstanceStateAnnotationName] = ErrorState
+
 		if err := oc.client.Update(nil, machine); err != nil {
 			return fmt.Errorf("unable to update machine status: %v", err)
 		}
@@ -389,6 +482,8 @@ func (oc *OpenstackClient) updateAnnotation(machine *machinev1.Machine, id strin
 		return err
 	}
 	machine.ObjectMeta.Annotations[openstack.OpenstackIPAnnotationKey] = ip
+	machine.ObjectMeta.Annotations[MachineInstanceStateAnnotationName] = instance.Status
+
 	if err := oc.client.Update(nil, machine); err != nil {
 		return err
 	}
@@ -404,6 +499,11 @@ func (oc *OpenstackClient) updateAnnotation(machine *machinev1.Machine, id strin
 		Address: machine.Name,
 	})
 
+	networkAddresses = append(networkAddresses, corev1.NodeAddress{
+		Type:    corev1.NodeInternalDNS,
+		Address: machine.Name,
+	})
+
 	machineCopy := machine.DeepCopy()
 	machineCopy.Status.Addresses = networkAddresses
 
@@ -423,14 +523,13 @@ func (oc *OpenstackClient) requiresUpdate(a *machinev1.Machine, b *machinev1.Mac
 	// Do not want status changes. Do want changes that impact machine provisioning
 	return !reflect.DeepEqual(a.Spec.ObjectMeta, b.Spec.ObjectMeta) ||
 		!reflect.DeepEqual(a.Spec.ProviderSpec, b.Spec.ProviderSpec) ||
-		!reflect.DeepEqual(a.Spec.Versions, b.Spec.Versions) ||
 		a.ObjectMeta.Name != b.ObjectMeta.Name
 }
 
 func (oc *OpenstackClient) instanceExists(machine *machinev1.Machine) (instance *clients.Instance, err error) {
 	machineSpec, err := openstackconfigv1.MachineSpecFromProviderSpec(machine.Spec.ProviderSpec)
 	if err != nil {
-		return nil, err
+		return nil, fmt.Errorf("\nError getting the machine spec from the provider spec (machine/actuator.go 457): %v", err)
 	}
 	opts := &clients.InstanceListOpts{
 		Name:   machine.Name,
@@ -440,12 +539,12 @@ func (oc *OpenstackClient) instanceExists(machine *machinev1.Machine) (instance
 
 	machineService, err := clients.NewInstanceServiceFromMachine(oc.params.KubeClient, machine)
 	if err != nil {
-		return nil, err
+		return nil, fmt.Errorf("\nError getting a new instance service from the machine (machine/actuator.go 467): %v", err)
 	}
 
 	instanceList, err := machineService.GetInstanceList(opts)
 	if err != nil {
-		return nil, err
+		return nil, fmt.Errorf("\nError listing the instances (machine/actuator.go 472): %v", err)
 	}
 	if len(instanceList) == 0 {
 		return nil, nil
@@ -459,7 +558,7 @@ func (oc *OpenstackClient) createBootstrapToken() (string, error) {
 		return "", err
 	}
 
-	expiration := time.Now().UTC().Add(TokenTTL)
+	expiration := time.Now().UTC().Add(options.TokenTTL)
 	tokenSecret, err := bootstrap.GenerateTokenSecret(token, expiration)
 	if err != nil {
 		panic(fmt.Sprintf("unable to create token. there might be a bug somwhere: %v", err))
